﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System;
using System.Linq;

public class Square : FlatLandObject
{

    public LineRenderer pathRenderer; // About drawing paths.
    public List<Vector2> pathList = new List<Vector2>();
    public UIManager uiManager;

    public List<FlatLandObject> allObjects = new List<FlatLandObject>(); // Evety FlatLandObject is automatically added with Awake().
    public FlatLandObject attatchedObject = null;

    private Vector2 _currentPathEnd;

    // Start is called before the first frame update
    void Start()
    {
    }

    // Update is called once per frame
    public void Update()
    {
        if(Input.GetMouseButtonDown(1)) // If right mouse button is clicked
        {
            Ray ray = Camera.main.ScreenPointToRay(Input.mousePosition);
            if (Physics.Raycast(ray, out RaycastHit hit)) // If the click was on the background
            {
                if (Input.GetKey(KeyCode.LeftShift) && pathRenderer.GetPosition(1) != Vector3.zero) // If shift is clicked and path is already exists
                    AddPath(hit.point);
                else
                    CreatePath(hit.point);
            }
        }

        if (Input.GetKeyDown(KeyCode.G))
            Grab();

        if (Input.GetKeyDown(KeyCode.H))
            Detach();
    }

    public IEnumerator MoveSquare(Vector2 destination)
    {
        speedVector = destination - (Vector2)transform.position;
        Vector3 scaledVector = speedVector * Time.deltaTime;
        while (speedVector.x * (destination.x - transform.position.x) > 0 || speedVector.y * (destination.y - transform.position.y) > 0)
        {
            transform.position += scaledVector;
            if (attatchedObject != null)
                attatchedObject.gameObject.transform.position += scaledVector;
            yield return null;
        }

    }

    public void CreatePath(Vector3 point) // Creates the fitst path, and updates Proper Time UI.
    {
        pathRenderer.positionCount = 2;
        pathRenderer.SetPositions(new Vector3[] { transform.position, point });
        uiManager.UpdateUI();
    }

    public void AddPath(Vector3 point) // Adds new path to current path, and updates Proper Time UI.
    {
        pathRenderer.positionCount++;
        pathRenderer.SetPosition(pathRenderer.positionCount - 1, point);
        uiManager.UpdateUI();
    }

    public void Move()
    {
        StartCoroutine(_StartMovingPath());
    }


    public IEnumerator _StartMovingPath()
    {
        Vector3[] tomoveList = new Vector3[pathRenderer.positionCount];
        pathRenderer.GetPositions(tomoveList);
        for(int i = 1; i < pathRenderer.positionCount; i++)
        {
            yield return StartCoroutine(MoveSquare(tomoveList[i]));
        }

        AddTimes();
        ResetPaths();
        uiManager.UpdateUI();
        uiManager.CheckVictory();
    }

    public void ResetPaths()
    {
        pathRenderer.positionCount = 2;
        pathRenderer.SetPositions(new Vector3[] { Vector3.zero, Vector3.zero });
    }

    public Vector2 getNthPath(int n) // returns movement vector stored in path renderer by index.
    {
        if (n >= pathRenderer.positionCount)
            throw new InvalidOperationException(n + "th path is not stored.");

        return pathRenderer.GetPosition(n + 1) - pathRenderer.GetPosition(n);
    }

    public double MovingTime(Vector2 v) // How long it takes to move given vector with current speed.
    {
        return v.sqrMagnitude / gameSpeed;
    }

    public double CalculateEntireMovingTime() // How long it takes to move current path all the way.
    {
        double result = 0;
        for (int i = 0; i < pathRenderer.positionCount - 1; i++)
            result += MovingTime(getNthPath(i));

        return result / gameSpeed;
    }

    public void ModifySpeed(float d)
    {
        if (gameSpeed + d <= Constants.c && gameSpeed + d >= 0)
            gameSpeed += d;
        uiManager.UpdateUI();
    }

    public void Grab()
    {
        var lst = Physics.OverlapSphere(transform.position, 10);

        if (lst.Length > 0)
            attatchedObject = lst[0].gameObject.GetComponent<FlatLandObject>();
    }

    public void Detach()
    {
        attatchedObject = null;
    }

    public void AddTimes()
    {
        Debug.Log(CalculateEntireMovingTime());
        foreach (var v in allObjects)
        {
            Debug.Log(v.gameObject.name);
            v.properTime += CalculateEntireMovingTime();
        }
        square.properTime -= CalculateEntireMovingTime() - (square.CalculateEntireMovingTime() / Constants.Gamma(square.gameSpeed));

        if (attatchedObject != null)
            attatchedObject.properTime -= CalculateEntireMovingTime() - (square.CalculateEntireMovingTime() / Constants.Gamma(square.gameSpeed));
    }
}
